"""
generators/feature.py
Ian Kollipara
2022.11.25

Feature Generation
"""

# Imports
from os import curdir
from pathlib import Path
from cstack._internal import templater
from tomllib import loads

PROJECT = loads(Path("pyproject.toml").read_text("utf-8"))


class FeatureGenerator:
    def __init__(self, feature_name: str) -> None:
        self.feature_path = (
            Path(curdir)
            / PROJECT["tool"]["poetry"]["name"]
            / Path("api")
            / "features"
            / feature_name
        )

    def create_router(self):
        (self.feature_path / "router.py").write_text(
            templater.render(
                "feature_router.py.j2", {"feature_name": self.feature_path.name.title()}
            )
        )

    def create_subdirectories(self):
        self.feature_path.mkdir()
        for directory in ("DTOs", "services", "repositories", "repositories/mock"):
            (self.feature_path / directory).mkdir()

    def create_feature_barrel(self):
        (self.feature_path / "__init__.py").write_text(
            templater.render(
                "feature_init.py.j2", {"feature_name": self.feature_path.name.title()}
            )
        )

    def create_test_file(self):
        (
            Path(curdir)
            / PROJECT["tool"]["poetry"]["name"]
            / Path("api")
            / "tests"
            / f"test_{self.feature_path.name.lower()}.py"
        ).write_text(
            templater.render(
                "feature_test.py.j2", {"feature_name": self.feature_path.name.title()}
            )
        )



"""
templater.py
Ian Kollipara
2022.11.25

Jinja2 Templater
"""

from typing import Any
from jinja2 import Environment, PackageLoader

_environment = Environment(loader=PackageLoader("cstack"))


def render(template_name: str, template_args: dict[str, Any]):
    return _environment.get_template(template_name).render(template_args)



"""
generators/model.py
Ian Kollipara
2022.11.26

Tortoise Model Generator
"""

# Imports
from os import curdir
from pathlib import Path
from tomllib import loads
from cstack._internal import templater

PROJECT = loads(Path("pyproject.toml").read_text("utf-8"))


class ModelGenerator:
    def __init__(self, model_name: str) -> None:
        self.model_path = (
            Path(curdir)
            / PROJECT["tool"]["poetry"]["name"]
            / Path("api")
            / "models"
            / f"{model_name.lower()}.py"
        )

    def __call__(self):
        self.model_path.write_text(
            templater.render(
                "api_model.py.j2",
                {"model_name": self.model_path.name.removesuffix(".py").title()},
            )
        )

        with (
            Path(curdir)
            / PROJECT["tool"]["poetry"]["name"]
            / Path("api")
            / "models"
            / "__init__.py"
        ).open("a") as f:
            f.write(
                f"from .{self.model_path.name.removesuffix('.py').lower()} import {self.model_path.name.removesuffix('.py').title()}"
            )



"""
core/application.py
Ian Kollipara
2022.11.25

Cstack Main Application Configuration
"""

# Imports
from pathlib import Path
from tomllib import loads
from fastapi import FastAPI
from .api import API
from .spa import SinglePageApplication

PROJECT = loads(Path("pyproject.toml").read_text("utf-8"))


def Application(*, api: API, view_dist: str = "view/dist"):
    api = api
    view = SinglePageApplication(
        directory=f"{PROJECT['tool']['poetry']['name']}/{view_dist}"
    )

    app = FastAPI()
    app.mount("/api", api.app, "api")
    app.mount("/", view, "view")
    return app



"""
core/api.py
Ian Kollipara
2022.11.25

API Class Definition
"""

# Imports
from fastapi import FastAPI, APIRouter


class API:
    def __init__(self) -> None:
        self.app = FastAPI()

    def include_routers(self, routes: dict[str, APIRouter]):
        for path, router in routes.items():
            self.app.include_router(router, prefix=path)



"""
generators/project.py
Ian Kollipara
2022.11.25

Project Generator
"""

# Imports
from pathlib import Path
from os import system, chdir, curdir, getcwd
from typing import Any, Literal
from cstack._internal import templater


class ProjectGenerator:
    def __init__(self, project_name: str) -> None:
        self.home_path = Path(project_name.lower())
        self.proj_path = Path(project_name.lower()) / Path(
            project_name.replace(" ", "_").replace("-", "_").lower()
        )

    def create_project_directory(
        self, use_vue: bool = True, node_manager: Literal["npm", "yarn", "pnpm"] = "npm"
    ):
        self.home_path.mkdir()
        self.proj_path.mkdir()
        (self.proj_path / "api").mkdir()
        (self.proj_path / "__init__.py").write_text(
            templater.render(
                "__init__.py.j2", {"project_name": self.proj_path.name.title()}
            )
        )
        if use_vue:
            match node_manager:
                case "npm":
                    system(f"npm create vue@3 {self.proj_path}/view")
                case "yarn":
                    system(f"yarn create vue {self.proj_path}/view")
                case "pnpm":
                    system(f"pnpm create vue {self.proj_path}/view")

    def create_api_subdirectory(self, dirname: str):
        (self.proj_path / "api" / dirname).mkdir()

    def scaffold_api_file(
        self, filename: str, template: str, template_args: dict[str, Any]
    ):
        (self.proj_path / "api" / f"{filename}").write_text(
            templater.render(template, template_args)
        )

    def scaffold_top_level_file(
        self, filename: str, template: str, template_args: dict[str, Any]
    ):
        (self.home_path / f"{filename}").write_text(
            templater.render(template, template_args)
        )

    def install(self):
        chdir(self.home_path.relative_to(curdir))
        system("poetry install")



"""
Cstack CLI 
"""

# Imports
from pathlib import Path
from typer import Typer, Option
from rich.console import Console
from rich.prompt import Confirm
from enum import Enum
from cstack.cli.generators.project import ProjectGenerator
from cstack import __version__


app = Typer()
console = Console()


class NodeManagerChoices(str, Enum):
    NPM = "npm"
    YARN = "yarn"
    PNPM = "pnpm"


@app.command()
def new(
    project_name: str,
    dry_run: bool = Option(
        False, "--dry-run", help="Show output but don't create any files"
    ),
    no_vue: bool = Option(False, "--no-vue", help="Don't create frontend with Vue"),
    node_manager: NodeManagerChoices = Option(
        NodeManagerChoices.NPM, "--using", help="Which package manager to use."
    ),
):

    project = ProjectGenerator(project_name)

    if not dry_run:
        project.create_project_directory(not no_vue, node_manager.value)

    console.print(
        f":snake: [blue on white reverse] CREATED [/blue on white reverse] {project.proj_path.name.title()} Directory"
    )
    if not no_vue:
        console.print(
            f":computer: [blue on white reverse] CREATED [/blue on white reverse][green bold] Vue Frontend "
        )

    for directory in ("models", "features", "tests", "config"):
        if not dry_run:
            project.create_api_subdirectory(directory)
        console.print(
            f":sparkles: [blue on white reverse] CREATED [/blue on white reverse] Models Directory"
        )

    for file_config in (
        (Path("config") / "settings.py", "api_config_settings.py.j2", {}),
        (Path("config") / "env.py", "api_config_env.py.j2", {}),
        (
            Path("config") / "db.py",
            "api_config_db.py.j2",
            {"project_name": project.proj_path.name.title()},
        ),
        ("app.py", "api_app.py.j2", {"project_name": project.proj_path.name.title()}),
    ):
        if not dry_run:
            project.scaffold_api_file(*file_config)
        console.print(
            f":sparkles: [blue on white reverse] CREATED [/blue on white reverse] {file_config[0]}"
        )

    for file_config in (
        (".env", "env_file.j2", {}),
        (
            "pyproject.toml",
            "pyproject.toml",
            {"project_name": project.proj_path.name.lower(), "version": __version__},
        ),
        ("cstack", "cstack", {}),
    ):
        if not dry_run:
            project.scaffold_top_level_file(*file_config)
        console.print(
            f":sparkles: [blue on white reverse] CREATED [/blue on white reverse] {file_config[0]}"
        )

    if Confirm.ask(
        "Run [black on white reverse] Poetry Install [/black on white reverse]?"
    ):
        project.install()

    console.print(
        f":grey_exclamation: [bold blue] Creation Finished! [/bold blue]\nTo Continue type:\n`$ cd {project.proj_path.name}`"
    )



"""
generators/service.py
Ian Kollipara
2022.11.25

Service Generator
"""

# Imports
from os import curdir
from tomllib import loads
from typing import Literal
from cstack._internal import templater
from pathlib import Path

PROJECT = loads(Path("pyproject.toml").read_text("utf-8"))


class ServiceGenerator:
    def __init__(self, name: str, feature_name: str) -> None:
        self.service_name = (
            Path(curdir)
            / PROJECT["tool"]["poetry"]["name"]
            / Path("api")
            / "features"
            / feature_name
            / f"{name.lower()}.py"
        )

    def __call__(self, type_: Literal["module", "class"]):
        match type_:
            case "module":
                self.service_name.write_text(
                    templater.render(
                        "service_module.py.j2",
                        {"service_name": self.service_name.name.title()},
                    )
                )
            case "class":
                self.service_name.write_text(
                    templater.render(
                        "service_class.py.j2",
                        {"service_name": self.service_name.name.title()},
                    )
                )



"""
core/middleware.py
Ian Kollipara
2022.11.25

Middleware Closure
"""

# Imports
from typing import Literal
from fastapi import FastAPI


def middleware(middleware_type: Literal["http", "ws"] = "http"):
    def inner(func):
        def wrapped(app: FastAPI):

            return app.middleware(middleware_type)(func)

        return wrapped

    return inner



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cstack',
 'cstack._internal',
 'cstack.cli',
 'cstack.cli.generators',
 'cstack.core']

package_data = \
{'': ['*'], 'cstack': ['templates/*']}

install_requires = \
['Jinja2>=3.1.2,<4.0.0',
 'aiofiles>=22.1.0,<23.0.0',
 'fastapi>=0.87.0,<0.88.0',
 'pydantic[dotenv]>=1.10.2,<2.0.0',
 'rich>=12.6.0,<13.0.0',
 'tortoise-orm>=0.19.2,<0.20.0',
 'typer>=0.7.0,<0.8.0',
 'uvicorn>=0.20.0,<0.21.0']

entry_points = \
{'console_scripts': ['cstack = cstack.cli.app:app']}

setup_kwargs = {
    'name': 'cstack',
    'version': '0.1.2',
    'description': 'Fullstack Python Framework. Built on Python, Vite, and Vue.',
    'long_description': "# 🐍 Cstack - Python made Easy\nAn Opinionated Full Stack Framework for Web Development. Built on Python, Vite, and Vue. \n\nThis stack can be used to create a fullstack application that can be deployed on Deta, or another python hosting.\nIt includes:\n- Monadic Error Handling\n- ORM\n- Testing Library\n- CLI Generator\n- Type Safety\n\n## Project Structure\n```\napp/\n | api/                  -- Python Project\n    | app.py             -- API Entrypoint\n    | models/            -- Database ORM Models\n    | config/            -- Pydantic Configuration\n       | db.py           -- Database Configuration\n       | env.py          -- Environment Variables\n       | settings.py     -- Application Settings\n    | features/          -- Feature Libraries\n    | shared/            -- Shared Code\n    | tests/             -- Pytest Tests\n | view/                 -- Vue Project\n __init__.py             -- Application File\n.env                     -- Env File\ncstack                   -- CLI Tool\npyproject.toml           -- Poetry File\n```\n\n\n### Features\n\nThis is the lifeblood of a Cstack project.\nCstack is feature-driven, rather than layered.\nThe reason is feature code is co-located, rather than spread across directories.\nA feature includes:\n- `__init__.py`\n- `router.py`\n- `services/`\n- `repositories/`\n- `repositories/mocks/`\n- `DTOs/`\n\n#### `router.py`\n\nThis is the fastapi router.\nAll endpoints for this function should be defined here.\n\n#### `services/`\n\nThis is collection of services.\nThese services can be defined as a either a class or as a collection of functions.\nThey should be pure operations, and contain most of the business logic.\nThe reason for purity is testability and portability.\nIf for a reason you move away from Cstack, the code is easily usable again.\nIt also means that the feature logic could be shared if needed.\n\nA service can be generated as well, these are injected into routes through FastAPI's dependency injection.\nA service includes:\n- `{{service_name}}.py`\n- `test_{{service_name}}.py`\n\n#### `repositories/`\n\nA repository communicates with the database. The reason for both repositories and an ORM is it allows the business logic\nto be completely decoupled from the database pattern. This is also means the services are easily testable.\nA repository includes:\n- `{{repository_name}}.py`\n- `mocks/{{repository_name}}.py`\n\nThe mocks folder contains dummy implementations of the repo, for use in testing.\n\n#### `DTOs/`\n\nThis a collection of Pydantic Objects.\nThese are used for communicating data for the given feature.\n\n> ORM Models ARE NOT DTOs. Use a Pydantic Model. Leave ORM code in the Repositories.\n\n### `config/db.py`\n\nAll the database setup code is contained here.\nIt is exposed to the user if the need arises to modify it.\nAll models should import the declared Database base.\n\n### `config/settings.py`\n\nThis includes versioning info, as well as other global settings used in the project.\nThis is instantiated as a singleton, and should only be used through that singleton.\n\n### `config/env.py`\n\nThis is influenced by T3-stack's env folder, where they define a runtime schema for Environment Variables.\nThese variables are defined here, and parsed through Pydantic.\nThis is instantiated as a singleton and should only be used through that singleton.",
    'author': 'Ian Kollipara',
    'author_email': 'ian.kollipara@cune.org',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.11,<4.0',
}


setup(**setup_kwargs)
